//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Top module for the SD card controller
// In SP605 board it interfaces with System ACE chip
// via MPU port
//
`include "sd_parms.v"
//
module sdcard_cntr  (
  clk_i, rst_i,  
  // MPU interface
  mpbrdy_i, mpirq_i,
  mpce_n_o, mpwe_n_o, mpoe_n_o, mpa_o,
  mpd_io, 
  // Wishbone interface
  wb_adr_i, wb_data_i, wb_we_i, wb_sel_i, wb_stb_i, wb_ack_i,
  wb_cyc_i, wb_tagn_i,
  wb_adr_o, wb_data_o, wb_we_o, wb_sel_o, wb_stb_o, wb_ack_o,
  wb_cyc_o, wb_tagn_o,
 // Test I/O
  led_o, test1_o, test2_o
  );
  
  input clk_i;                   // Input clock
  input rst_i;                   // Reset signal
  input mpbrdy_i;                // Buffer ready
  input mpirq_i;                 // Interrupt request
  output mpce_n_o;               // Chip enable
  output mpwe_n_o;               // Write enable
  output mpoe_n_o;               // Output enable
  output [6:0] mpa_o;            // Address
  inout [7:0] mpd_io;            // Data lines
  // Wishbone interface
  input [31:0] wb_adr_i;         // Input address
  input [31:0] wb_data_i;        // Input data
  input wb_we_i;                 // Write enable signal
  input [3:0] wb_sel_i;          // Byte select signal
  input wb_stb_i;                // Input strobe
  input wb_ack_i;                // Input acknowledge signal
  input wb_cyc_i;                // Input cycle signal
  input wb_tagn_i;               // Input tag signal
  output [31:0] wb_adr_o;        // Output address
  output [31:0] wb_data_o;       // Output data
  output wb_we_o;                // Output write enable signal
  output [3:0] wb_sel_o;         // Output select signal
  output wb_stb_o;               // Output strobe signal
  output wb_ack_o;               // Output acknowledge signal
  output wb_cyc_o;               // Output cycle signal
  output wb_tagn_o;              // Output tag signal
  // Test I/O
  output [3:0] led_o;            // Output to LEDs
  output [3:0] test1_o;          // Test output
  output [15:0] test2_o;         // Test output
  
  reg mpce_n_o;
  reg mpwe_n_o;
  reg mpoe_n_o;
  reg [6:0] mpa_o;
  
  // Internal wires and registers
  wire [3:0] sd_fsm_w;
  reg [15:0] bus_counter;
  reg bus_cycle_done;
  reg dwrite_r;
  
  reg [7:0] mpd_drv;  
  reg [7:0] mpd_buf;
  reg [31:0] mpd_buf32;
  reg [6:0] mpa_buf;
  wire [3:0] cmd_w;
  wire [31:0] data_w;
  reg [15:0] wcnt_r;
  reg next_state1_r;
  reg next_state2_r;
  reg [31:0] data_buf [0:7];      // Internal buffer to keep one ACE buffer - 32 bytes
  reg [2:0] buf_adr;
  reg [15:0] buf_cnt;             // Count of buffers
  reg buf_er_r;
  reg [3:0] buf_count;
  reg [15:0] to_count;            // Timeout counter
  reg flip_bit;
  reg timeout;
  reg [2:0] byte_cnt; 
  reg [15:0] test_cnt = 16'h0;
  
  reg data_read_r;
  reg xfer_done_lock;
  wire xfer_done_w;
  wire data_avail_w;  
  wire [3:0] wb_test_w;
  
  assign mpd_io = (~mpwe_n_o) ? mpd_drv : 8'hz; 
  assign led_o = sd_fsm_w;						
  assign test2_o = test_cnt;
  
  // Modules instantiation
  wb_cntr_ace wb_cntr_sd  (
    .clk_i              (clk_i),
    .rst_i              (rst_i),
    .mpd_i              (mpd_buf32),
    .wcnt_i             (wcnt_r),
    .dwrite_i           (dwrite_r),
    .data_read_i        (data_read_r),
    .timeout_i          (timeout),
    .cmd_o              (cmd_w),
    .data_o             (data_w),
    .data_avail_o       (data_avail_w),
    .xfer_done_o        (xfer_done_w),
    .wb_adr_i           (wb_adr_i), 
    .wb_data_i          (wb_data_i), 
    .wb_we_i            (wb_we_i), 
    .wb_sel_i           (wb_sel_i), 
    .wb_stb_i           (wb_stb_i), 
    .wb_ack_i           (wb_ack_i),
    .wb_cyc_i           (wb_cyc_i), 
    .wb_tagn_i          (wb_tagn_i),
    .wb_adr_o           (wb_adr_o), 
    .wb_data_o          (wb_data_o), 
    .wb_we_o            (wb_we_o), 
    .wb_sel_o           (wb_sel_o), 
    .wb_stb_o           (wb_stb_o), 
    .wb_ack_o           (wb_ack_o),
    .wb_cyc_o           (wb_cyc_o), 
    .wb_tagn_o          (wb_tagn_o),
    .test_o             (wb_test_w)
  );
  
  sdcard_fsm sdcard_fsm_sd (
    .clk_i              (clk_i), 
    .rst_i              (rst_i), 
    .cmd_i              (cmd_w),
    .bus_cycle_done_i   (bus_cycle_done),
    .next_state1_i      (next_state1_r),
    .next_state2_i      (next_state2_r),
    .buf_er_i           (buf_er_r),
    .sd_fsm_o           (sd_fsm_w)
  );
  
  
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          mpce_n_o <= 1'b1;
          mpwe_n_o <= 1'b1;
          mpoe_n_o <= 1'b1;
          dwrite_r <= 1'b0;
        end
      else
        begin
          case (sd_fsm_w)           // Synthesis full case
            `SD_IDLE:
              begin
                bus_counter <= 16'h0;
                to_count <= 16'h0;
                bus_cycle_done <= 1'b0;
                dwrite_r <= 1'b0;
                next_state1_r <= 1'b0; 
                next_state2_r <= 1'b0;
                buf_cnt <= 16'h0;
                buf_er_r <= 1'b0;
                data_read_r <= 1'b0;
                mpa_o <= (`BUF_S - 1);      //Init required by SD_WDB state
                xfer_done_lock <= 1'b0;
                timeout <= 1'b0;
              end
            `SD_SRR:               //Signle register read
              begin
                bus_counter <= bus_counter + 1'b1;
                wcnt_r <= 16'h1;
                if (bus_counter == 16'h0)
                  begin
                    mpa_o <= data_w[6:0];
                    mpce_n_o <= 1'b0;
                    mpwe_n_o <= 1'b1;
                    mpoe_n_o <= 1'b1;
                  end
                else if (bus_counter == 16'h2)
                  begin
                    mpoe_n_o <= 1'b0;
                  end
                else if (bus_counter == 16'h3)
                  begin
                    mpoe_n_o <= 1'b1;
                    mpce_n_o <= 1'b1;
                    bus_cycle_done <= 1'b1;
                    bus_counter <= 16'h0;
                    mpd_buf <= mpd_io;
                  end					
              end
            `SD_SRR_WB:
              begin
                dwrite_r <= 1'b1;
                mpd_buf32 <= {24'h0,mpd_buf};
               if (wb_ack_i)
                 begin
                   next_state1_r <= 1'b1;  
                   dwrite_r <= 1'b0;
                 end
              end
            `SD_SRW:                // Single register write
              begin
                bus_counter <= bus_counter + 1'b1;
                if (bus_counter == 16'h0)
                  begin
                    mpa_o <= data_w[6:0];
                    mpce_n_o <= 1'b0;
                    mpwe_n_o <= 1'b1;
                    mpoe_n_o <= 1'b1;
                  end				 
                else if (bus_counter == 16'h1)
                  begin
                    mpwe_n_o <= 1'b0;
                    mpd_drv <= data_w[15:8];
                  end						
                else if (bus_counter == 16'h2)
                  begin
                    mpwe_n_o <= 1'b1;	   
                    mpce_n_o <= 1'b1;
                    bus_cycle_done <= 1'b1;
                    bus_counter <= 16'h0;
                  end
              end
            `SD_MRR:
              begin
                bus_counter <= bus_counter + 1'b1;
                wcnt_r <= 16'h1;
                if (bus_counter == 16'h0)
                  begin
                    mpa_o <= data_w[6:0];
                    mpce_n_o <= 1'b0;
                    mpwe_n_o <= 1'b1;
                    mpoe_n_o <= 1'b1;
                    mpd_buf32 <= 32'h0;
                  end
                else if (bus_counter == 16'h2)
                  begin
                    mpoe_n_o <= 1'b0;
                  end
                else if (bus_counter == 16'h3)
                  begin
                    mpoe_n_o <= 1'b1;
                    mpd_buf32[7:0] <= mpd_io;            // First register recorded
                  end															   
                else if ((bus_counter == 16'h4) || (bus_counter == 16'h9) || (bus_counter == 16'hE))
                  begin
                    mpa_o <= mpa_o + 1'b1;
                  end					  
                else if ((bus_counter == 16'h7) || (bus_counter == 16'hC) || (bus_counter == 16'h11))
                  begin
                     mpoe_n_o <= 1'b0;
                  end				  
                else if ((bus_counter == 16'h8) || (bus_counter == 16'hD))
                  begin
                    mpd_buf32 <= {mpd_buf32[23:0],mpd_io};
                    mpoe_n_o <= 1'b1;
                  end																				
                else if (bus_counter == 16'h12)                // Four byte cycle is completed
                  begin
                    mpd_buf32 <= {mpd_buf32[23:0],mpd_io};
                    mpoe_n_o <= 1'b1;
                    mpce_n_o <= 1'b1;
                    bus_counter <= 16'h0;
                    bus_cycle_done <= 1'b1;
                  end
              end
            `SD_MRR_WB:
              begin
                dwrite_r <= 1'b1;
               if (wb_ack_i)
                 begin
                   next_state1_r <= 1'b1;  
                   dwrite_r <= 1'b0;
                 end
              end
            `SD_MRW:
              begin
                bus_counter <= bus_counter + 1'b1;	
                if (bus_counter == 16'h0)
                  begin
                    mpa_o <= data_w[6:0];
                    mpce_n_o <= 1'b0;
                    mpwe_n_o <= 1'b1;
                    mpoe_n_o <= 1'b1;
                  end				
                else if (bus_counter == 16'h2)
                  begin
                    mpwe_n_o <= 1'b0;
                    mpd_drv <= data_w[31:24];
                    mpd_buf32 <= data_w;
                  end						 
                else if ((bus_counter == 16'h3) || (bus_counter == 16'h5) || (bus_counter == 16'h7))
                  begin
                    mpa_o <= mpa_o + 1'b1;
                    mpwe_n_o <= 1'b1;
                    mpd_buf32 <= {mpd_buf32[23:0],8'h0};
                  end					
                else if ((bus_counter == 16'h4) || (bus_counter == 16'h6) || (bus_counter == 16'h8))
                  begin	                     
                    mpwe_n_o <= 1'b0;
                    mpd_drv <= mpd_buf32[31:24];
                  end																			   
                else if (bus_counter == 16'h9)                // Write cycle completed
                  begin
                    mpwe_n_o <= 1'b1;	
                    mpce_n_o <= 1'b1;
                    bus_cycle_done <= 1'b1;
                    bus_counter <= 16'h0;
                  end
              end
            `SD_RDB:
              begin
                next_state2_r <= 1'b0;
                if (bus_counter == 16'h7)
                  bus_counter <= 16'h2;
                else if ((bus_counter == 16'h1) && ~mpbrdy_i)
                  to_count <= to_count + 1'b1;	
                else if ((bus_counter == 16'h2) && ~mpbrdy_i)
                  to_count <= to_count + 1'b1;
                else
                  bus_counter <= bus_counter + 1'b1;
                
                if (to_count == `TO_RDB)
                  begin
                    buf_er_r <= 1'b1;
                    wcnt_r <= 16'h1;
                  end
                  
                if (bus_counter == 16'h0)
                  begin
                    buf_cnt <= data_w[15:0];
                  end
                else if (bus_counter == 16'h1)
                  begin
                    bus_cycle_done <= 1'b0;
                    mpa_o <= `BUF_S;                
                    buf_adr <= 3'b000;
                    mpce_n_o <= 1'b0;
                    mpwe_n_o <= 1'b1; 
                    mpoe_n_o <= 1'b1;    
                    byte_cnt <= 2'b00;
                  end
                else if (bus_counter == 16'h2)
                  begin
                    mpoe_n_o <= 1'b1;
                  end
                else if (bus_counter == 16'h4)
                  begin				  
                    mpoe_n_o <= 1'b0;
                  end
                else if (bus_counter == 16'h5)
                  begin
                    mpoe_n_o <= 1'b1;
                    mpa_o[0] <= ~mpa_o[0];        // Flip bit zero to change between 40 and 41
                    mpd_buf32 <= {mpd_buf32[23:0],mpd_io};            // First register recorded
                    byte_cnt <= byte_cnt + 1'b1;
                  end
                else if (bus_counter == 16'h6)
                  begin
                    if (byte_cnt == 3'b100)    // One 32 bit word has been completed
                      begin
                        data_buf[buf_adr] <= mpd_buf32;
                        buf_adr <= buf_adr + 1'b1;
                        byte_cnt <= 3'b000;
                        if (buf_adr == 3'b111)
                          begin
                            buf_cnt <= buf_cnt - 1'b1;
                            buf_count <= 4'h1;
                            bus_cycle_done <= 1'b1;
                            mpd_buf32 <= data_buf[0];
                            wcnt_r <= 16'h8;
                          end
                      end
                  end
              end
            `SD_RDB_WB:       // Buffer of 8 32-bit words to be transmitted over WB
              begin			
                bus_cycle_done <= 1'b0;
                to_count <= 16'h0;
                dwrite_r <= 1'b1;     // Words count has been initialized in the SD_RDB state
                if (wb_ack_i)
                  begin
                    mpd_buf32 <= data_buf[buf_count[2:0]];
                    buf_count <= buf_count + 1'b1;
                    if (buf_count == 4'b1000)     // Exit this state
                      begin
                        dwrite_r <= 1'b0;
                        if (buf_cnt == 16'h0)       // All the data has been Tx'd - go to the IDLE state
                          begin
                            next_state1_r <= 1'b1;
                            mpce_n_o <= 1'b1;
                          end
                        else        // Still some more data to be Tx'd
                          begin
                            next_state2_r <= 1'b1;
                            bus_counter <= 16'h1;
                          end
                      end
                  end
              end
            `SD_RDB_ER:
              begin	 
                dwrite_r <= 1'b1;
                mpd_buf32 <= `ER_PATTERN;
                if (wb_ack_i)
                  begin
                    dwrite_r <= 1'b0;
                    next_state2_r <= 1'b1;
                  end
              end
            `SD_WDB:
              begin
                mpwe_n_o <= 1'b1;
                mpce_n_o <= 1'b0;
                if (xfer_done_w)
                  xfer_done_lock <= 1'b1;
                    
                if (data_avail_w && ~timeout)               // One 32-bit word has been acquired and needs to be written to the flash
                  begin
                    next_state1_r <= 1'b1;
                    data_read_r <= 1'b1;
                    bus_counter <= 16'h0;
                    to_count <= 16'h0;
                    flip_bit <= 1'b0;
                  end
                else if (xfer_done_lock)
                  begin
                    next_state2_r <= 1'b1;      // Write back status
                    wcnt_r <= 16'h1;
                    if (timeout)
                      mpd_buf32 <= `TO_PATTERN;
                    else
                      mpd_buf32 <= `C_PATTERN;
                  end
              end
            `SD_WDB_WM:
              begin
                data_read_r <= 1'b0;
                to_count <= to_count + 1'b1;
                
                if (xfer_done_w)
                  xfer_done_lock <= 1'b1;
                  
                if (to_count == 16'h0)
                  begin
                    mpd_buf32 <= data_w;
                    mpa_o <= `BUF_S;
                  end
                else if (mpbrdy_i && flip_bit)
                  mpd_buf32 <= {mpd_buf32[23:0],8'h0};
                  
                if (mpbrdy_i)
                  begin
                    if (~flip_bit)
                      begin
                        mpa_o[0] <= ~mpa_o[0];
                        mpwe_n_o <= 1'b1;
                        mpoe_n_o <= 1'b1;
                        flip_bit <= ~flip_bit;
                      end
                    else 
                      begin
                        mpwe_n_o <= 1'b0;
                        mpd_drv <= mpd_buf32[31:24];
                        flip_bit <= ~flip_bit;
                        bus_counter <= bus_counter + 1'b1;
                        if (bus_counter == 16'h3)
                          begin
                            next_state1_r <= 1'b0;        // Return to SD_WDB state
                            next_state2_r <= 1'b0;
                          end
                      end
                  end				
                else if (to_count == `TO_VALUE)       // Timeout occurred
                  begin
                    mpwe_n_o <= 1'b1;
                    mpoe_n_o <= 1'b1;
                    mpce_n_o <= 1'b1;
                    timeout <= 1'b1;
                    next_state1_r <= 1'b0;
                    next_state2_r <= 1'b0;
                  end
              end
            `SD_WDB_WB:
              begin
                dwrite_r <= 1'b1;
                mpce_n_o <= 1'b1;
                if (wb_ack_i)
                  begin
                    dwrite_r <= 1'b0;
                    next_state2_r <= 1'b0;
                  end
              end
          endcase
        end
    end
 /*   
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
        
        end
      else
        begin
          case (sd_fsm_w)           // Synthesis full case
            `SD_SRR:
              begin
                if (bus_counter == 16'h4)
                  mpd_buf <= mpd_io;  
              end
          endcase
        end  
      
    end
  */  
endmodule