//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Wishbone controller for SD controller
//
// It decodes commands for SD controller
// `BASE_ADDRESS - read one register. Format of data: {25'h0,7'h[ADR]} - first seven bits represent register to be read
// `BASE_ADDRESS + 1 - write one register. Format of data: {16'h0,8'h[DATA],1'b0,7'h[ADR]} - [DATA] to be recorded into register [ADR]  
// `BASE_ADDRESS + 2 - read multiple registers - four. Format of data: {16'h0,8'h0,1'b0,7'h[START_ADR]} - Command to read 4 registers starting
//                                                from START_ADR. Address incremented by 1
// `BASE_ADDRESS + 3 - write multiple registers - four. Format of data: {16'h0,8'h0,1'b0,7'h[START_ADR]} followed
//         by 32-bit word in format {R1,R2,R3,R4}. Address incremented by 1
// `BASE_ADDRESS + 4 - read data buffer. Format of data: {16'h0,5'h0,11'h[Number_of_buffers]}. Every buffer contains 32 bytes of data
//                  Maximum amount of data trasnferred per one command is 16Kb.
// `BASE_ADDRESS + 5 - write data buffer. Format of data: {16'h0,5'h0,11'h[Number_of_buffers]} followed by data Every buffer contains 32 bytes of data
//                  Maximum amount of data transferred per one command is 16Kb.
//
`include "sd_parms.v"

  module wb_cntr_ace (
    clk_i, rst_i, mpd_i, wcnt_i, dwrite_i, data_read_i, timeout_i,
    cmd_o, data_o, data_avail_o, xfer_done_o,
    // Wishbone interface
    wb_adr_i, wb_data_i, wb_we_i, wb_sel_i, wb_stb_i, wb_ack_i,
    wb_cyc_i, wb_tagn_i,
    wb_adr_o, wb_data_o, wb_we_o, wb_sel_o, wb_stb_o, wb_ack_o,
    wb_cyc_o, wb_tagn_o,
    // Test I/O
    test_o
    );
    
    input clk_i;                  // Input clock 
    input rst_i;                  // Input reset signal
    input [31:0] mpd_i;           // Input data from the ACE chip      
    input [15:0] wcnt_i;          // Number of words to be transmitted
    input dwrite_i;               // Request to write data over WB from internal consumers
    input data_read_i;            // Read data from internal buffer
    input timeout_i;              // Signal that timeout occured during writing buffer
    output [3:0] cmd_o;           // Decoded command
    output [31:0] data_o;         // Data from WB for internal consumers
    output data_avail_o;          // Signal that data is available to be read - asserted on the negedge
    output xfer_done_o;           // Signal that data buffer receipt has been completed
    input [31:0] wb_adr_i;        // Input address
    input [31:0] wb_data_i;       // Input data
    input wb_we_i;                // Write enable
    input [3:0] wb_sel_i;         // Byte select
    input wb_stb_i;               // Strobe
    input wb_ack_i;               // Acknowledge
    input wb_cyc_i;               // WB cycle
    input wb_tagn_i;              // Tag
    output [31:0] wb_adr_o;       // Output address
    output [31:0] wb_data_o;      // Output data
    output wb_we_o;               // Write enable
    output [3:0] wb_sel_o;        // Byte select
    output wb_stb_o;              // Strobe
    output wb_ack_o;              // Acknowledge
    output wb_cyc_o;              // Bus cycle indicator
    output wb_tagn_o;             // Tag
    output [3:0] test_o;          // Test output
    
    reg [31:0] wb_adr_o;
    reg [31:0] wb_data_o;
    reg wb_we_o;
    reg wb_stb_o;
    reg wb_ack_o;
    reg wb_cyc_o;
    reg wb_tagn_o;
    reg [3:0] wb_sel_o;
    reg [3:0] cmd_o;
    reg [31:0] data_o;
    reg xfer_done_o;
    
    // Internal registers and wires
  
    reg next_state1;  
    reg next_state2;
    reg [7:0] counter;
    reg wb_ack_pos;
    reg [15:0] wcnt_buf;
    reg [31:0] data_buf;
    reg [15:0] buf_cnt;
    reg [31:0] wb_buf [0:7];        // Memory to store one buffer
    reg [3:0] adr_wr;               // Address to write to the buffer with one extra bit
    reg [3:0] adr_rd;               // Address to read from the buffer with one extra bit
    reg data_avail;
    reg data_read_neg;
    
    reg [3:0] wbsd_fsm_state;
    reg [3:0] wbsd_fsm_next_state;
    
    assign data_avail_o = data_avail;			
    assign test_o = wbsd_fsm_state;
    
    //-----------WB bus state machine---------------------
    always @ (wbsd_fsm_state or wb_we_i or wb_cyc_i or wb_stb_i or next_state1
              or dwrite_i or next_state2)
      begin
        wbsd_fsm_next_state = `WBSD_IDLE;
        case (wbsd_fsm_state)                   // Synthesis full case
          `WBSD_IDLE:
            begin
              if (dwrite_i)
                wbsd_fsm_next_state = `WBSD_BWRITE;
              else if (wb_we_i && wb_cyc_i && wb_stb_i)
                begin
                  wbsd_fsm_next_state = `WBSD_DECODE;
                end  
              else
                wbsd_fsm_next_state = `WBSD_IDLE;
            end
          `WBSD_DECODE:
            begin
              if (next_state1 && ~next_state2)
                wbsd_fsm_next_state = `WBSD_IDLE;
              else if (next_state2 && ~next_state1)
                wbsd_fsm_next_state = `WBSD_WREAD;
              else if (next_state1 && next_state2)
                wbsd_fsm_next_state = `WBSD_BUF_READ;
              else						
                wbsd_fsm_next_state = `WBSD_DECODE;
            end
          `WBSD_BWRITE:
            begin
              if (next_state1)
                wbsd_fsm_next_state = `WBSD_IDLE;
              else
                wbsd_fsm_next_state = `WBSD_BWRITE;
            end									  
          `WBSD_WREAD:						  
            begin							
              if (next_state1)
                wbsd_fsm_next_state = `WBSD_WREAD_T;
              else
                wbsd_fsm_next_state = `WBSD_WREAD;
            end 
          `WBSD_WREAD_T:
            begin						
              if (next_state2)
                wbsd_fsm_next_state = `WBSD_IDLE;
              else
                wbsd_fsm_next_state = `WBSD_WREAD_T;
            end
          `WBSD_BUF_READ:
            begin
              if (next_state1 && ~next_state2)
                wbsd_fsm_next_state = `WBSD_IDLE;
              else
                wbsd_fsm_next_state = `WBSD_BUF_READ;
            end
        endcase
      end
    
    always @ (negedge clk_i)
      begin
        if (!rst_i)
          begin
            wbsd_fsm_state <= `WBSD_IDLE;
          end
        else
          begin
            wbsd_fsm_state <= wbsd_fsm_next_state;
          end
      end
    
    //---------End of WB bus state machine----------------
    
    always @ (posedge clk_i)
      begin
        if (!rst_i)
          begin
            wb_we_o <= 1'b0;
            wb_stb_o <= 1'b0;
            wb_ack_pos <= 1'b0;
            wb_cyc_o <= 1'b0;
            next_state1 <= 1'b0;
            next_state2 <= 1'b0;
          end
        else
          begin
            case (wbsd_fsm_state)                   // Synthesis full case
              `WBSD_IDLE:
                begin
                  next_state1 <= 1'b0;
                  next_state2 <= 1'b0;
                  counter <= 8'h0;
                  wcnt_buf <= wcnt_i;
                  xfer_done_o <= 1'b0;
                end
              `WBSD_DECODE:
                begin
                  counter <= counter + 1'b1;
                  if (counter == 8'h0)
                    begin 
                      if (wb_adr_i == `WBSD_BASE)
                        begin
                          cmd_o <= `CMD_SRR;
                          data_o <= wb_data_i;
                          wb_ack_pos <= 1'b1;
                        end 
                      else if (wb_adr_i == (`WBSD_BASE + 1))
                        begin
                          cmd_o <= `CMD_SRW;
                          data_o <= wb_data_i;
                          wb_ack_pos <= 1'b1;
                        end 
                      else if (wb_adr_i == (`WBSD_BASE + 2))
                        begin
                          cmd_o <= `CMD_MRR;
                          data_o <= wb_data_i;
                          wb_ack_pos <= 1'b1;
                        end
                      else if (wb_adr_i == (`WBSD_BASE + 3))
                        begin                               
                          data_o <= wb_data_i;
                          wb_ack_pos <= 1'b1;
                        end
                      else if (wb_adr_i == (`WBSD_BASE + 4))
                        begin
                          cmd_o <= `CMD_RDB;
                          data_o <= wb_data_i;
                          wb_ack_pos <= 1'b1;
                        end
                      else if (wb_adr_i == (`WBSD_BASE + 5))
                        begin
                          cmd_o <= `CMD_WDB;
                          data_o <= wb_data_i;
                          buf_cnt <= (wb_data_i[15:0] << 3);  // Get dwords count from buffer count
                          wb_ack_pos <= 1'b1;
                          adr_wr <= 4'h0;
                          adr_rd <= 4'h0;
                        end
                    end
                  else if (counter == 8'h1)
                    begin
                      if (wb_adr_i == (`WBSD_BASE + 3))
                        begin
                          next_state2 <= 1'b1;
                          next_state1 <= 1'b0;
                        end
                      else if (wb_adr_i == (`WBSD_BASE + 5))
                        begin
                          next_state2 <= 1'b1;
                          next_state1 <= 1'b1;
                        end
                      else
                        begin
                          next_state1 <= 1'b1;
                          next_state2 <= 1'b0;
                        end
                      wb_ack_pos <= 1'b0;
                      cmd_o <= `CMD_IDLE;
                    end
                end
              `WBSD_BWRITE:
                begin
                  wb_we_o <= 1'b1;
                  wb_stb_o <= 1'b1;
                  wb_cyc_o <= 1'b1;
                  wb_sel_o <= 4'hF;
                  wb_data_o <= mpd_i;
                  if (wb_ack_i)
                    wcnt_buf <= wcnt_buf - 1'b1;
                  if (wcnt_buf == 16'h0)
                    begin
                      next_state1 <= 1'b1;
                      wb_we_o <= 1'b0;
                      wb_stb_o <= 1'b0;
                      wb_cyc_o <= 1'b0;
                    end
                end		
              `WBSD_WREAD:
                begin
                   counter <= 8'h0;
                   if (wb_we_i && wb_cyc_i && wb_stb_i)
                     begin
                       data_buf <= wb_data_i;
                       cmd_o <= `CMD_MRW;
                       wb_ack_pos <= 1'b1;		 
                       next_state1 <= 1'b1;
                       next_state2 <= 1'b0;
                     end
                end
              `WBSD_WREAD_T:
                begin	
                  counter <= counter + 1'b1;
                  wb_ack_pos <= 1'b0;
                  cmd_o <= `CMD_IDLE;
                  if (counter == 8'h1)
                    begin		   
                      data_o <= data_buf;
                      next_state2 <= 1'b1;
                    end
                end
              `WBSD_BUF_READ:
                begin
                  counter <= 8'h0;
                  data_o <= wb_buf[adr_rd[2:0]];
                  if (wb_we_i && wb_cyc_i && wb_stb_i && ~wb_ack_o)
                    begin
                      if (adr_wr[3:0] != {~adr_rd[3],adr_rd[2:0]})
                        begin
                          wb_buf[adr_wr[2:0]] <= wb_data_i;
                          adr_wr <= adr_wr + 1'b1;
                          wb_ack_pos <= 1'b1;
                          buf_cnt <= buf_cnt - 1'b1;
                        end
                      else if (timeout_i)     // If timeout occured confirm reception regardless of the buffer availability
                        begin
                          buf_cnt <= buf_cnt - 1'b1;
                          wb_ack_pos <= 1'b1;
                        end
                    end
                    
                  if (data_read_neg)
                    adr_rd <= adr_rd + 1'b1;
                      
                  if (wb_ack_pos)
                    begin
                      wb_ack_pos <= 1'b0;
                    end
                  if ((buf_cnt == 16'h0) && (data_read_neg | timeout_i))
                    begin
                      next_state2 <= 1'b0;        // Go back to the IDLE state
                      xfer_done_o <= 1'b1;
                    end
                end
            endcase
          end
      end
      
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          data_avail <= 1'b0;
        end
      else
        begin
          if (adr_wr != adr_rd)
            data_avail <= 1'b1;
          else
            data_avail <= 1'b0;
        end
      wb_ack_o <= wb_ack_pos;
      data_read_neg <= data_read_i;
    end    
      
 endmodule
 